/*
 *    Copyright (C) 2002 Sergej Bogomolov <bog@deol.ru>
 *    This filter convert from raw planar 4:2:2 to YUV4MPEG2
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <config.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <yuv4mpeg.h>
#include <mjpeg_logging.h>

int	verbose = 1;
uint8_t	*input_frame[3];
uint8_t	*output_frame[3];
int	*stab_offset;

int	interlace=0;
int	width=768;
int	height=576;
int	deinterlace=0;
int	luma_threshold=16;
int	stab_limit=0;

static void Usage(char *name )
{
	fprintf(stderr,
		"Usage: %s: [-?] [-w width] [-h height] [-i interlace] [-d] [-s threshold] [-v num]\n"
                        "-?   - print out this help\n"
			"-w   - frame width (default: 768)\n"
			"-h   - frame height (default: 576)\n"
			"-i   - interlace in input (0 - none, 1 - top, 2 - bottom)\n"
			"-d   - deinterlace to 50 Hz\n"
			"-s   - line stabilisation (luma coring threshold: 16)\n"
			"-v   - Verbosity [0..2]\n", name);
}

size_t readbytes( int fd, void *buf, size_t count ) {
	ssize_t status;
	int rem = count;
	void *address=buf;
	while( rem > 0 ) {
		status = read( fd, address, rem );
		if( status <= 0 ) {
			fprintf( stderr, "read status: %d\n", status );
			return status;
		}
		rem -= status;
		address += status;
	}
	return count;
}

void filter_color (uint8_t *input, uint8_t * const output) {
	int i, j, w2=width/2;
	uint8_t *in=input, *out=output;
	if( interlace ) {
		for (i=0; i<height; i+=2, in += width, out += w2) {
			for( j=0; j<w2; j++) out[j] = (in[j] + in[j+width])/2;
		}
	} else {
		for (i=0; i<height; i+=2, in += width, out += w2) {
			for( j=0; j<w2; j++) out[j] = (in[j] + in[j+w2])/2;
		}
	}
}

void stab_check( uint8_t *input ) {
	int i, j;
	uint8_t *in=input;
	if( stab_limit) {
	  for( i=0;i<height;i++, in += width ) {
	  	stab_offset[i] = stab_limit;
		for( j=stab_limit;j>=0;j--)
			if( in[j] > luma_threshold ) stab_offset[i] = j;
		if(stab_offset[i] == stab_limit ) stab_offset[i] = 0;
	  }
	} else for( i=0;i<height;i++ ) stab_offset[i] = 0;
}

void filter(uint8_t *input[], uint8_t * const output[], int offset )
{
	int i, j, w2=width/2;
	uint8_t *in, *out;
	stab_check( input[0] );
	if( deinterlace ) {
		in=input[0]+offset*width;
		out=output[0];
		for (i=offset; i<height; i+=2, in += width*2, out += width*2) {
			memcpy( out, in+stab_offset[i], width );
			memcpy( out+width, in+stab_offset[i], width );
		}
		in=input[1]+offset*w2;
		out=output[1];
		for (i=offset; i<height; i+=2, in += width, out += w2) {
			for( j=0; j<w2; j++) out[j] = (in[j] + in[j+width])/2;
		}
		in=input[2]+offset*w2;
		out=output[2];
		for (i=offset; i<height; i+=2, in += width, out += w2) {
			for( j=0; j<w2; j++) out[j] = (in[j] + in[j+width])/2;
		}
	} else {
		memcpy( output[0], input[0], width*height );
		filter_color( input[1], output[1]);
		filter_color( input[2], output[2]);
	}
}

int main(int argc, char *argv[])
{
	int	input_fd = 0;
	int	output_fd = 1;
	int	c;
	int	frame_count;
	int i;
	int yymin=256, yymax=-1, uumin=256, uumax=-1, vvmin=256, vvmax=-1;
	long yhist[256] = {256*0}, uhist[256] = {256*0}, vhist[256] = {256*0};

	y4m_stream_info_t ostream;
	y4m_frame_info_t oframe;

	while((c = getopt(argc, argv, "w:h:v:i:s:d?")) != EOF) {
		switch(c) {
			case 'w':
				width = atoi(optarg);
				break;
			case 'h':
				height = atoi(optarg);
				break;
			case 'i':
				interlace = atoi(optarg);
				break;
			case 'd':
				deinterlace = 1;
				break;
			case 's':
				luma_threshold = atoi(optarg);
				stab_limit = 100;
				break;
		case 'v':
			verbose = atoi (optarg);
			if( verbose < 0 || verbose >2 )
			{
				Usage (argv[0]);
				exit (1);
			}
			break;		  
			
		case '?':
                        Usage (argv[0]);
		default:
			exit(0);
		}
	}

	(void)mjpeg_default_handler_verbosity( verbose );

	y4m_init_stream_info( &ostream );
	y4m_init_frame_info( &oframe );

	y4m_si_set_width( &ostream, width);
	y4m_si_set_height( &ostream, height);
	
	if( deinterlace ) {
		y4m_si_set_interlace( &ostream, Y4M_ILACE_NONE );
		y4m_si_set_framerate( &ostream, y4m_fps_PAL_FIELD );
	} else {
		switch( interlace ) {
		case '1':
			y4m_si_set_interlace( &ostream, Y4M_ILACE_TOP_FIRST );
			break;
		case '2':
			y4m_si_set_interlace( &ostream, Y4M_ILACE_BOTTOM_FIRST);
			break;
		default:
			y4m_si_set_interlace( &ostream, Y4M_ILACE_NONE );
		}
		y4m_si_set_framerate( &ostream, y4m_fps_PAL );
	}
	y4m_si_set_sampleaspect( &ostream, y4m_sar_UNKNOWN );
	
	if( ostream.height % 4 != 0 )
	{
		mjpeg_error_exit1("Input images have odd number of lines" );
	}
	if( ostream.width % 2 != 0 )
	{
		mjpeg_error_exit1("Input images have odd number of pixels" );
	}

	mjpeg_debug( "width=%d height=%d interlace=%d", 
			width, height, interlace);

	input_frame[0] = malloc(width * height);
	if(!input_frame[0]) { perror( "malloc" ); exit(1); }
	input_frame[1] = malloc((width / 2) * height);
	if(!input_frame[1]) { perror( "malloc" ); exit(1); }
	input_frame[2] = malloc((width / 2) * height);
	if(!input_frame[2]) { perror( "malloc" ); exit(1); }

	output_frame[0] = malloc(width * height);
	if(!output_frame[0]) { perror( "malloc" ); exit(1); }
	output_frame[1] = malloc((width / 2) * (height / 2));
	if(!output_frame[1]) { perror( "malloc" ); exit(1); }
	output_frame[2] = malloc((width / 2) * (height / 2));
	if(!output_frame[2]) { perror( "malloc" ); exit(1); }

	stab_offset = malloc( height );
	if(!stab_offset) { perror( "malloc" ); exit(1); }

	y4m_write_stream_header(output_fd, &ostream);
fprintf( stderr, "before loop\n" );
	frame_count = 0;
	for (;;)
	{ 
		int ymin=256, ymax=-1, umin=256, umax=-1, vmin=256, vmax=-1;

		i = readbytes( input_fd, input_frame[0], width*height );
		if( i != width*height ) break;
		if(verbose >=2 ) for( i=0; i<width*height; i++ ) {
			if( input_frame[0][i] < ymin ) ymin = input_frame[0][i];
			if( input_frame[0][i] > ymax ) ymax = input_frame[0][i];
			yhist[input_frame[0][i]]++;
		}
		if( ymin < yymin ) yymin = ymin;
		if( ymax > yymax ) yymax = ymax;
		i = readbytes( input_fd, input_frame[1], width*height / 2 );
		if( i != width*height / 2 ) break;
		if(verbose >=2 ) for( i=0; i<width*height/2; i++ ) {
			if( input_frame[1][i] < umin ) umin = input_frame[1][i];
			if( input_frame[1][i] > umax ) umax = input_frame[1][i];
			uhist[input_frame[1][i]]++;
		}
		if( umin < uumin ) uumin = umin;
		if( umax > uumax ) uumax = umax;
		i = readbytes( input_fd, input_frame[2], width*height / 2 );
		if( i != width*height / 2 ) break;
		if(verbose >=2 ) for( i=0; i<width*height/2; i++ ) {
			if( input_frame[2][i] < vmin ) vmin = input_frame[2][i];
			if( input_frame[2][i] > vmax ) vmax = input_frame[2][i];
			vhist[input_frame[2][i]]++;
		}
		if( vmin < vvmin ) vvmin = vmin;
		if( vmax > vvmax ) vvmax = vmax;
		filter( input_frame, output_frame, 0 );
		y4m_write_frame(output_fd, &ostream, &oframe, output_frame);
		if( deinterlace ) {
		  filter( input_frame, output_frame, 1 );
		  y4m_write_frame(output_fd, &ostream, &oframe, output_frame);
		}
		mjpeg_debug( "frame=%d, Y': %d-%d, Cb: %d-%d, Cr:%d-%d", 
			frame_count, ymin, ymax, umin, umax, vmin, vmax );
		frame_count++;
	}
	mjpeg_debug( "total Y': %d-%d, Cb: %d-%d, Cr:%d-%d", 
				yymin, yymax, uumin, uumax, vvmin, vvmax );
	if( verbose >= 2 ) {
		double	mhist, chist;
		fprintf( stderr, "\nY' histogramm:\n\n" );
		mhist=-1.;
		for( i=yymin; i<=yymax;  i++ ) {
		  chist = (double)yhist[i]/((double)width*height*frame_count);
		  if( mhist < chist ) mhist = chist;
		}
		for( i=yymin; i<=yymax;  i++ ) {
		  chist = (double)yhist[i]/((double)width*height*frame_count);
		  fprintf( stderr, "  %3d: %f %-60.*s\n", i, chist, (int)(chist/mhist*60), "************************************************************" );
		}
		fprintf( stderr, "\nCb histogramm:\n\n" );
		mhist=-1.;
		for( i=uumin; i<=uumax;  i++ ) {
		 chist = (double)uhist[i]/((double)width/2.*height*frame_count);
		 if( mhist < chist ) mhist = chist;
		}
		for( i=uumin; i<=uumax;  i++ ) {
		 chist = (double)uhist[i]/((double)width/2.*height*frame_count);
		 fprintf( stderr, "  %4d: %f %-60.*s\n", i-128, chist, (int)(chist/mhist*60), "************************************************************" );
		}
		fprintf( stderr, "\nCr histogramm:\n\n" );
		mhist=-1.;
		for( i=vvmin; i<=vvmax;  i++ ) {
		 chist = (double)vhist[i]/((double)width/2.*height*frame_count);
		 if( mhist < chist ) mhist = chist;
		}
		for( i=vvmin; i<=vvmax;  i++ ) {
		 chist = (double)vhist[i]/((double)width/2.*height*frame_count);
		 fprintf( stderr, "  %4d: %f %-60.*s\n", i-128, chist, (int)(chist/mhist*60), "************************************************************" );
		}
	}
	y4m_fini_stream_info(&ostream);
	y4m_fini_frame_info(&oframe);
	exit(0);
}
